# DupeFinder.py
# main application object for DupeFinder
#
# Copyright 2004 Matt Heinzen

# This file is part of DupeFinder 1.0.
#
# DupeFinder is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# DupeFinder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with File Duplicate Finder; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# TODO: handle issues where duplicate files are moved, renamed or deleted
# externally, probably a combination of refreshing the view periodically and
# verifying that a file exists when an operation is performed on it
# TODO: store a dictionary of file extensions to viewer commands
# TODO: show progress dialog for duplicate finding?

# REPORTED BUGS AND SUGGESTIONS
# Whenever the filename contains characters like ~ or : I am getting a "no such
#   file or directory". Also, whenever a symlink points to itself ( ln -s /test
#   test ) DupeFinder seems to die.
# 0 byte files are listed fine and everything, but whenever I select them for
#   deletion and want to delete them, everything stays as it is ... nothing
#   happens.
# whenever a filename contains characters like the german "umlauts" (����
#   it will be displayed fine on the gui, but deletion will fail and the program
#   will not continue. One has to de-select the file causing problems and the
#   deletion works.
# whenever one hits the "uncheck all marks" button and there are more than e.g.
#   20 or 50 files selected, a requester with "Are you sure?" would be great ...
#   when I had a few hundred selected I had it hit by mistake instead of the
#   "delete marked files" button and had to re-select everything.
# cannot handle broken symbolic links

import FindDupFiles
from FindDupMainUI import FindDupMainUI
from FindDupResultUI import FindDupResultUI

import sys, os
import string, re
from qt import *


markedDelete=[
	"40 14 3 1",
	" 	c None",
	"x	c black",
	"o	c red",
	"                                        ",
	"                 oo          ooo        ",
	"                oooo       ooo          ",
	"                ooooo     ooo           ",
	"                 ooooo   ooo            ",
	"                  ooooo ooo             ",
	"                    oooooo              ",
	"                      oooo              ",
	"                     oooooo             ",
	"                    ooo oooo            ",
	"                   ooo    ooo           ",
	"                  ooo      ooo          ",
	"                 ooo         oo         ",
	"                oo             o        ",
	]

markedClear=[
	"16 14 2 1",
	" 	c None",
	".	c black",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	"                ",
	]



class MainWin(FindDupMainUI):
	def __init__(self, *args):
		FindDupMainUI.__init__(self, *args)

		self.connect(self.chooseDirBtn, SIGNAL("clicked()"),
			self.chooseDir)
		self.connect(self.findDupsBtn, SIGNAL("clicked()"),
			self.findDups)
		self.dirEdit.setText(os.getcwd())

		# create an empty list of results windows
		self.resultWins = []


	def chooseDir(self):
		# show a QFileDialog to allow selection of the search directory
		dirStr = QFileDialog.getExistingDirectory(".", self,
			"choose search directory", "Choose the top search directory", 1)
		if len(str(dirStr)) > 0:
			self.dirEdit.setText(dirStr)


	def findDups(self):
		# display a wait cursor for the application
		QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

		# find all duplicate files
		dups = FindDupFiles.findDups(str(self.dirEdit.text()),
			FindDupFiles.fileExtsFromString(str(self.extEdit.text())))

		# restore the normal cursor
		QApplication.restoreOverrideCursor()

		# create a new window containing the results
		if len(dups) > 0:
			resultWin = ResultWin()
			resultWin.initDupTable(dups)
			resultWin.show()

			# keep a reference to the dialog so it is not GC'd when we leave
			# the local scope
			# NOTE: we use a list here so that an unlimited of different results
			# windows can be open at once. This is a little inelegant because
			# the list does not shrink as result windows are closed. A
			# dictionary which has references removed when windows are closed
			# might be more efficient, but because the list is never parsed and
			# will typically stay small, it should do fine
			self.resultWins.append(resultWin)
		else:
			QMessageBox.information(self, "DupeFinder",
				"No duplicates found!")



class ResultWin(FindDupResultUI):
	def __init__(self, *args):
		FindDupResultUI.__init__(self, *args)

		self.connect(self.viewSelFileBtn, SIGNAL("clicked()"),
			self.viewSelFile)
		self.connect(self.renameSelFileBtn, SIGNAL("clicked()"),
			self.renameSelFile)
		self.connect(self.saveReportBtn, SIGNAL("clicked()"), self.saveReport)
		self.connect(self.allowSelAllCheck, SIGNAL("clicked()"),
			self.toggleAllowGroupDelete)
		self.connect(self.markAllFilesBtn, SIGNAL("clicked()"),
			self.markAllFiles)
		self.connect(self.clearAllMarksBtn, SIGNAL("clicked()"),
			self.unmarkAllFiles)
		self.connect(self.deleteMarkedBtn, SIGNAL("clicked()"),
			self.deleteAllMarked)
		self.connect(self.dupsListView,
			SIGNAL("clicked(QListViewItem *, const QPoint &, int)"),
			self.viewClicked)
		self.connect(self.dupsListView,
			SIGNAL("doubleClicked(QListViewItem *, const QPoint &, int)"),
			self.viewDoubleClicked)
		self.connect(self.dupsListView, SIGNAL("spacePressed(QListViewItem *)"),
			self.spacePressed)
		self.connect(self.dupsListView, SIGNAL("selectionChanged()"),
			self.updateButtons)

		# initialize bookkeeping counters used to enable/disable buttons
		self.nTotalGroups = 0
		self.nTotalFiles = 0
		self.nMarkedToDelete = 0
		self.nFullyMarkedGroups = 0

		# disable buttons
		self.viewSelFileBtn.setEnabled(0)
		self.renameSelFileBtn.setEnabled(0)
		self.markAllFilesBtn.setEnabled(self.allowSelAllCheck.isChecked())
		self.clearAllMarksBtn.setEnabled(0)
		self.deleteMarkedBtn.setEnabled(0)

		# init listview "checkbox" text values
		self.deleteText = " "
		self.clearText = ""

		# create pixmaps for items marked and not marked for deletion
		self.deletePixmap = QPixmap(markedDelete)
		self.clearPixmap = QPixmap(markedClear)

		# init empty shell command and arguments for launching an external
		# viewer
		self.shellCmd = ""
		self.shellArgs = []


	def initDupTable(self, dups):
		# setup the table columns
		while self.dupsListView.columns() > 0:
			self.dupsListView.removeColumn(0)
		self.dupsListView.addColumn("Size")
		self.dupsListView.addColumn("Delete?")
		self.dupsListView.addColumn("File Path")

		# show list item selection in all columns
		self.dupsListView.setAllColumnsShowFocus(1)

		# show an indicator for the column used for sorting
		self.dupsListView.setShowSortIndicator(1)

		# show "+/-" symbols for top-level (group) items in tree
		self.dupsListView.setRootIsDecorated(1)

		# insert the data, leaving all items unselected by default
		if dups != None:
			for group in dups:
				size = str(group[0])
				size = " " * (16 - len(size)) + size;
				groupItem = QListViewItem(self.dupsListView, size)
				self.dupsListView.insertItem(groupItem)
				groupItem.setOpen(1)
				for path in group[2]:
					childItem = QListViewItem(groupItem, "", "", path)
				self.nTotalGroups += 1
				self.nTotalFiles += len(group[2])

		# store the duplicates object
		self.dups = dups


	def numMarked(self, groupItem):
		childItem = groupItem.firstChild()
		n = 0
		while childItem:
			if childItem.text(1) != self.clearText: n += 1
			childItem = childItem.nextSibling()
		return n


	def areAnyMarked(self, groupItem):
		childItem = groupItem.firstChild()
		while childItem:
			if childItem.text(1) != self.clearText: return 1
			childItem = childItem.nextSibling()
		return 0


	def areAllMarked(self, groupItem):
		childItem = groupItem.firstChild()
		while childItem:
			if childItem.text(1) == self.clearText: return 0
			childItem = childItem.nextSibling()
		return 1


	def updateButtons(self):
		item = self.dupsListView.selectedItem()
		self.viewSelFileBtn.setEnabled(item != None and item.childCount() == 0)
		self.renameSelFileBtn.setEnabled(item != None
			and item.childCount() == 0)
		self.markAllFilesBtn.setEnabled(
			self.allowSelAllCheck.isChecked()
			and self.nMarkedToDelete < self.nTotalFiles)
		self.clearAllMarksBtn.setEnabled(self.nMarkedToDelete > 0)
		self.deleteMarkedBtn.setEnabled(self.nMarkedToDelete > 0)
		self.allowSelAllCheck.setEnabled(self.nFullyMarkedGroups == 0)


	def queryShellCmd(self):
		# prompt user for a shell command and parse into separate arguments
		cmd = QInputDialog.getText("Start Viewer",
			"Enter the shell command to start a viewer for the selected file:",
				QLineEdit.Normal, string.join([self.shellCmd] + self.shellArgs))
		tokens = re.split(" ", str(cmd[0].simplifyWhiteSpace()))

		# do not store command if none is specified to prevent deletion of
		# previous (in case user canceled input dialog rather than entering a
		# command)
		if len(tokens[0]) == 0:
			return 0

		# store shell command and option arguments
		self.shellCmd = tokens[0]
		self.shellArgs = tokens[1:]
		return 1


	def execShellCmd(self, filePath):
		if not os.path.exists(filePath):
			QMessageBox.warning(self, "Error!",
				"The specified file no longer exists\n"
				"at the previous location!")
			return

		os.spawnvp(os.P_NOWAIT, self.shellCmd, [self.shellCmd] + self.shellArgs
			+ [filePath])


	def toggleItemDelete(self, item):
		# toggle clicked item status and update marked item counter
		if item.text(1) == self.clearText:
			if (self.allowSelAllCheck.isChecked()
				or self.numMarked(item.parent())
				< (item.parent().childCount() - 1)):

				# mark the Delete? column and update the marked group and file
				# counters
				item.setText(1, self.deleteText)
				item.setPixmap(1, self.deletePixmap)
				self.nMarkedToDelete += 1
				if self.areAllMarked(item.parent()):
					self.nFullyMarkedGroups += 1
			else:
				# display error message indicating entire groups cannot
				# currently be marked
				QMessageBox.warning(self, "Warning!",
					"<qt>Cannot mark every file in a duplicate "
					"group for deletion unless the <b>Allow deletion "
					"of all duplicates in a single group</b> "
					"box is checked.</qt>")
		else:
			# unmark the Delete? column and update the marked group and file
			# counters
			if self.areAllMarked(item.parent()):
				self.nFullyMarkedGroups -= 1
			item.setText(1, self.clearText)
			item.setPixmap(1, self.clearPixmap)
			self.nMarkedToDelete -= 1


	def viewClicked(self, item, point, column):
		if item != None and column == 1 and item.childCount() == 0:
			# toggle clicked item status and update marked item counter
			self.toggleItemDelete(item)

		# update GUI
		self.updateButtons()


	def viewDoubleClicked(self, item, point, column):
		if item != None and column == 2:
			# no previous command, prompt for shell command to execute viewer
			if self.shellCmd == "":
				if not self.queryShellCmd():
					return

			# shell command specified
			if self.shellCmd != "":
				self.execShellCmd(str(item.text(2)))


	def spacePressed(self, item):
		if item != None and item.childCount() == 0:
			self.toggleItemDelete(item)
			self.updateButtons()


 	def viewSelFile(self):
		# get the current filepath
		filePath = str(self.dupsListView.selectedItem().text(2))

		if self.queryShellCmd():
			self.execShellCmd(filePath)


	def renameSelFile(self):
		# get the current filepath
		oldPath = str(self.dupsListView.selectedItem().text(2))

		# query the user for a new file name and/or location
		newPath = str(QFileDialog.getSaveFileName(oldPath, ".", self,
			"report save dlg", "Save to New Name/Location"))
		if newPath != "":
			# abort if path is not changed
			if newPath == oldPath:
				QMessageBox.warning(self, "Same Name!",
					"The file name and directory specified is the same.\n"
					"Please choose a different file name or location when "
					"renaming files.")
				return
			# abort if file already exists
			elif os.path.exists(newPath):
				QMessageBox.warning(self, "File Already Exists!",
					"A file with the chosen name already exists.\n"
					"Please choose a unique file name when renaming files.")
				return

			# move the file on disk
			try:
				os.rename(oldPath, newPath)
			except:
				if not os.path.exists(oldPath):
					QMessageBox.warning(self, "Error!",
						"The specified file no longer exists\n"
						"at the previous location!")
				else:
					QMessageBox.warning(self, "Error!",
						"The file could not be moved.\nProbable reasons include "
						"incorrect permissions to\ncreate or modify files at the "
						"specified location.")
					return

			# update the display
			self.dupsListView.selectedItem().setText(2, newPath)

			# update the duplicates object
			size = int(str(self.dupsListView.selectedItem().parent().text(0)))
			for group in self.dups:
				if (group[0] == size):
					paths = group[2]
					for i in xrange(len(paths)):
						if paths[i] == oldPath:
							paths[i] = str(newPath)
							return


	def saveReport(self):
		# query the user for a report filename and open the file for writing
		filename = "dup_files"
		nameSet = 0
		while not nameSet:
			filename = QFileDialog.getSaveFileName(filename, ".", self,
				"report save dlg", "Save Report")

			# QFileDialog does not wanr users if they select an existing file,
			# so we popup our own warning here
			if os.path.exists(str(filename)):
				ans = QMessageBox.question(self, "File Exists",
					"<qt>The specified file<br><b>'"
					+ str(filename) + "'</b><br> already exists. "
					"Overwrite this file?</qt>", "Overwrite", "Choose New File",
					"Cancel", 0, 2)
				if ans == 0: nameSet = 1
				elif ans == 2: return
			else: nameSet = 1

		if len(filename) > 0:
			try:
				reportFile = open(str(filename), 'w')
			except:
				QMessageBox.warning(self, "Error!",
					"Could not create specified report file:\n"
					+ str(filename))
				return

			# create a dictionary mapping size to groups
			size2groups = {}
			for group in self.dups:
				size = group[0]
				if not size in size2groups:
					size2groups[size] = []
				size2groups[size].append(group)

			# extract and sort the list of group sizes
			sizes = size2groups.keys()
			sizes.sort()

			# write the duplicate data to the report file
			for size in sizes:
				for group in size2groups[size]:
					reportFile.write(str(size) + " bytes\n")
					for path in group[2]:
						reportFile.write(path + "\n")
					reportFile.write("\n")

			# close the report file
			reportFile.close()


	def toggleAllowGroupDelete(self):
		# simply update GUI to reflect new options
		self.updateButtons()


	def markAllFiles(self):
		# mark every file
		groupItem = self.dupsListView.firstChild()
		while groupItem:
			childItem = groupItem.firstChild()
			while childItem:
				childItem.setText(1, self.deleteText)
				childItem.setPixmap(1, self.deletePixmap)
				childItem = childItem.nextSibling()
			groupItem = groupItem.nextSibling()

		# reset marked file and group counter and update GUI
		self.nMarkedToDelete = self.nTotalFiles
		self.nFullyMarkedGroups = self.nTotalGroups
		self.updateButtons()


	def unmarkAllFiles(self):
		# unmark every file
		groupItem = self.dupsListView.firstChild()
		while groupItem:
			childItem = groupItem.firstChild()
			while childItem:
				childItem.setText(1, self.clearText)
				childItem.setPixmap(1, self.clearPixmap)
				childItem = childItem.nextSibling()
			groupItem = groupItem.nextSibling()

		# reset marked group and file counters and update GUI
		self.nFullyMarkedGroups = 0
		self.nMarkedToDelete = 0
		self.updateButtons()


	def deleteAllMarked(self):
		# start a dictionary of deleted files and a list of files which are not
		# deleted successfully
		removed = {}
		removeErrors = []

		# delete all marked files and remove any emptied dupe groups
		groupItem = self.dupsListView.firstChild()
		while groupItem:
			childItem = groupItem.firstChild()
			while childItem:
				nextItem = childItem.nextSibling()

				# delete the actual file and the tree view item
				if childItem.text(1) != "":
					removed[unicode(childItem.text(2))] = ""

					# delete the actual file
					try:
						os.remove(unicode(childItem.text(2)))
						groupItem.takeItem(childItem)
					# could not delete, save path for error report and unmark
					# path in view
					except:
						removeErrors.append(unicode(childItem.text(2)))
						childItem.setText(1, self.clearText)
						childItem.setPixmap(1, self.clearPixmap)

				childItem = nextItem

			nextItem = groupItem.nextSibling()
			if groupItem.childCount() == 0:
				self.dupsListView.takeItem(groupItem)
			groupItem = nextItem

		# some files were not deleted, display an error message
		if len(removeErrors):
			msgText = ""
			for error in removeErrors:
				msgText += error + "<br>"
			QMessageBox.warning(self, "Error!",
				"<qt>The following files could not be deleted:<br><b>"
				+ msgText + "</b>Probable reasons include incorrect permissions "
				"to<br>delete or modify files at the specified location(s).")

		# remove deleted files from duplicates object
		for group in self.dups:
			paths = group[2]
			for path in paths[:]:
				if path in removed:
					paths.remove(path)

		# remove empty dupe groups
		for group in self.dups[:]:
			if len(group[2]) == 0:
				self.dups.remove(group)

		# reset marked group and file counters and update GUI
		self.nFullyMarkedGroups = 0
		self.nMarkedToDelete = 0
		self.updateButtons()



def main(args):
	# create the Qt application object
	app = QApplication(args)

	# create and display the main GUI
	appUI = MainWin()
	appUI.show()

	# set the app to finish when the last window is closed
	app.connect(app, SIGNAL("lastWindowClosed()"), app,
		SLOT("quit()"))

	# run the app
	app.exec_loop()


if __name__=="__main__":
	main(sys.argv)
